// vim:fdm=syntax
// by tuberry
/* exported init */
'use strict';

const Main = imports.ui.main;
const Slider = imports.ui.slider;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;
const BoxPointer = imports.ui.boxpointer;
const Screenshot = imports.ui.screenshot;
const { Gio, St, Shell, GObject, Clutter, Meta } = imports.gi;

const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();
const { Fields } = Me.imports.fields;
const _ = ExtensionUtils.gettext;

const setCursor = cursor => global.display.set_cursor(Meta.Cursor[cursor]);

const Notify = { MSG: 0, OSD: 1 };
const Format = { HEX: 0, RGB: 1, HSL: 2, hex: 3, HSV: 4, CMYK: 5 };

class Color {
    constructor(text, format) {
        this._text = text || '#fff';
        this.format = format;
    }

    set format(format) {
        this._format = format;
    }

    get text_format() {
        return this._text_format ?? (this._text_format = this.toFormat(this._text));
    }

    get format() {
        return this._format ?? this.text_format;
    }

    toFormat(x) {
        if(x.startsWith('#')) return Format.HEX;
        else if(x.startsWith('rgb')) return Format.RGB;
        else if(x.startsWith('hsl')) return Format.HSL;
        else if(x.startsWith('hsv')) return Format.HSV;
        else if(x.startsWith('cmyk')) return Format.CMYK;
        else return Format.hex;
    }

    toText(format) {
        switch(format ?? this.format) {
        case Format.RGB: return (({ r, g, b }) => `rgb(${r}, ${g}, ${b})`)(this.rgb);
        case Format.HSL: return (({ h, s, l }) => `hsl(${Math.round(h)}, ${Math.round(s * 100)}%, ${Math.round(l * 100)}%)`)(this.hsl);
        case Format.HSV: return (({ h, s, v }) => `hsv(${Math.round(h)}, ${Math.round(s * 100)}%, ${Math.round(v * 100)}%)`)(this.hsv);
        case Format.CMYK: return (({ c, m, y, k }) => `cmyk(${c}, ${m}, ${y}, ${k})`)(this.cmyk);
        case Format.hex: return this.color.to_string().slice(1, 7);
        default: return this.color.to_string().slice(0, 7);
        }
    }

    toColor(t, f) {
        switch(f) {
        case Format.HEX:
        case Format.RGB:
            return Clutter.Color.from_string(t)[1];
        case Format.HSL: {
            let [h, s, l] = t.slice(4, -1).split(',').map((x, i) => parseInt(x) / (i ? 100 : 1));
            return Clutter.Color.from_hls(h, l, s);
        }
        case Format.HSV: {
            let [h, s, v] = t.slice(4, -1).split(',').map((x, i) => parseInt(x) / (i ? 100 : 1));
            let { h: hue, s: sat, l } = this.hsv2hsl({ h, s, v });
            return Clutter.Color.from_hls(hue, l, sat);
        }
        case Format.CMYK: {
            let [c, m, y, k] = t.slice(5, -1).split(',').map(v => parseInt(v));
            let { r, g, b } = this.cmyk2rgb({ c, m, y, k });
            return new Clutter.Color({ red: r, green: g, blue: b });
        }
        default: return Clutter.Color.from_string(`#${t}`)[1];
        }
    }

    get color() {
        return this._color ?? (this._color = this.toColor(this._text, this.text_format));
    }

    hsv2hsl({ h, s, v }) {
        let l = v * (1 - s / 2);
        let sl = l === 0 || l === 1 ? 0 : (v - l) / Math.min(l, 1 - l);
        return { h, l, s: sl };
    }

    // Refer: https://en.wikipedia.org/wiki/HSL_and_HSV
    hsl2hsv({ h, s, l }) {
        let v = l + s * Math.min(l, 1 - l);
        let sv = v === 0 ? 0 : 2 * (1 - l / v);
        return { h, s: sv, v };
    }

    get hsv() {
        return this.hsl2hsv(this.hsl);
    }

    cmyk2rgb({ c, m, y, k }) {
        [c, m, y, k] = [c, m, y, k].map(x => x / 255);
        let [r, g, b] = [c, m, y].map(x => Math.round((1 - x * (1 - k) - k) * 255));
        return { r, g, b };
    }

    // Refer: https://zh.wikipedia.org/wiki/%E5%8D%B0%E5%88%B7%E5%9B%9B%E5%88%86%E8%89%B2%E6%A8%A1%E5%BC%8F
    rgb2cmyk({ r, g, b }) {
        let cmy = [r, g, b].map(x => 1 - x / 255);
        let k = Math.min(...cmy);
        let [c, m, y, k1] = k === 1 ? [0, 0, 0, 1] : cmy.map(x => (x - k) / (1 - k)).concat(k).map(x => Math.round(x * 255));
        return { c, m, y, k: k1 };
    }

    get cmyk() {
        return this.rgb2cmyk(this.rgb);
    }

    set color(color) {
        this._color = color;
    }

    set rgb(color) {
        Object.assign(this._color, color);
    }

    set hsl(color) { // [h, l, s]
        let hls = this.color.to_hls();
        this._color = Clutter.Color.from_hls(...Object.assign(hls, color));
    }

    get hsl() {
        let [h, l, s] = this.color.to_hls();
        return { h, s, l };
    }

    get rgb() {
        let { red: r, green: g, blue: b } = this.color;
        return { r, g, b };
    }

    get markup() {
        let { l } = this.hsl;
        // NOTE: https://gitlab.gnome.org/GNOME/mutter/-/issues/1324
        return ` <span fgcolor="${Math.round(l) ? '#000' : '#fff'}" bgcolor="${this.toText(Format.HEX)}">${this.toText()}</span>`;
    }
}

class Field {
    constructor(prop, gset, obj) {
        this.gset = typeof gset === 'string' ? new Gio.Settings({ schema: gset }) : gset;
        this.prop = prop;
        this.bind(obj);
    }

    _get(x) {
        return this.gset[`get_${this.prop[x][1]}`](this.prop[x][0]);
    }

    _set(x, y) {
        this.gset[`set_${this.prop[x][1]}`](this.prop[x][0], y);
    }

    bind(a) {
        let fs = Object.entries(this.prop);
        fs.forEach(([x]) => { a[x] = this._get(x); });
        this.gset.connectObject(...fs.flatMap(([x, [y]]) => [`changed::${y}`, () => { a[x] = this._get(x); }]), a);
    }

    unbind(a) {
        this.gset.disconnectObject(a);
    }
}

class MenuItem extends PopupMenu.PopupMenuItem {
    static {
        GObject.registerClass(this);
    }

    constructor(text, callback, params) {
        super(text, params);
        this.connect('activate', callback);
    }

    setLabel(label) {
        if(this.label.text !== label) this.label.set_text(label);
    }
}

class RadioItem extends PopupMenu.PopupSubMenuMenuItem {
    static {
        GObject.registerClass(this);
    }

    constructor(name, modes, index, callback) {
        super('');
        this._name = name;
        this._list = Object.keys(modes);
        this._list.map((x, i) => new MenuItem(_(x), () => callback(i))).forEach(x => this.menu.addMenuItem(x));
        this.setSelected(index);
    }

    setSelected(index) {
        if(!(index in this._list)) return;
        this.label.set_text(`${this._name}：${this._list[index]}`);
        this.menu._getMenuItems().forEach((y, i) => y.setOrnament(index === i ? PopupMenu.Ornament.DOT : PopupMenu.Ornament.NONE));
    }
}

class IconItem extends PopupMenu.PopupBaseMenuItem {
    static {
        GObject.registerClass(this);
    }

    constructor(style, callbacks) {
        super({ activate: false });
        this._style = style;
        this._hbox = new St.BoxLayout({ x_align: St.Align.START, x_expand: true });
        callbacks.forEach(xs => this.addButton(...xs));
        this.add_child(this._hbox);
    }

    addButton(icon_name, callback) {
        let btn = new St.Button({ x_expand: true, style_class: this._style, child: new St.Icon({ icon_name, style_class: 'popup-menu-icon' }) });
        btn.connect('clicked', callback);
        this._hbox.add_child(btn);
    }
}

class ColorItem extends PopupMenu.PopupBaseMenuItem {
    static {
        GObject.registerClass(this);
    }

    constructor(color, style, callback) {
        super();
        this._label = new St.Label({ x_expand: true });
        this._button = new St.Button({ child: new St.Icon({ icon_name: 'emblem-favorite-symbolic', style_class: 'popup-menu-icon' }) });
        this._button.connect('clicked', () => this._call(this._color.text));
        [this._label, this._button].forEach(x => this.add_child(x));
        this.setLabel(color);
        this.setButton(style, callback);
        this.connect('activate', () => St.Clipboard.get_default().set_text(St.ClipboardType.CLIPBOARD, this._color.text));
    }

    setLabel(text) {
        if(this._color?.text === text) return;
        this._color = new Color(text);
        this._label.clutter_text.set_markup(this._color.markup);
    }

    setButton(style, callback) {
        this._call = callback;
        if(this._style === style) return;
        if(this._style) this._button.remove_style_class_name(this._style);
        this._button.add_style_class_name(this._style = style);
    }
}

class DListSection extends PopupMenu.PopupMenuSection {
    constructor(list, style, callback) {
        super();
        this.updateList(list, style, callback);
    }

    setList(list) {
        let items = this._items;
        let diff = list.length - items.length;
        if(diff > 0) for(let a = 0; a < diff; a++) this.addMenuItem(new ColorItem(null, this._style, this._call));
        else if(diff < 0) for(let a = 0; a > diff; a--) items.at(a - 1).destroy();
        this._items.forEach((x, i) => x.setLabel(list[i]));
    }

    updateList(list, style, callback) {
        this._style = style;
        this._call = callback;
        this.setList(list);
        this._items.forEach(x => x.setButton(this._style, this._call));
    }

    get _items() {
        return this._getMenuItems();
    }
}

class ColorSlider extends Slider.Slider {
    static {
        GObject.registerClass(this);
    }

    constructor(numb, base) {
        super(numb / base);
        this.base = base;
        this.step = base > 1 ? 1 / base : 0.01;
    }

    get numb() {
        return this.value * this.base;
    }

    set numb(numb) {
        this.value = numb / this.base;
    }

    vfunc_key_press_event(event) {
        let key = event.keyval;
        if(key === Clutter.KEY_Right || key === Clutter.KEY_Left) {
            let delta = key === Clutter.KEY_Right ? this.step : -this.step;
            this.value = Math.max(0, Math.min(this._value + delta, this._maxValue));
            return Clutter.EVENT_STOP;
        }

        return super.vfunc_key_press_event(event);
    }

    scroll(event) {
        if(event.is_pointer_emulated()) return Clutter.EVENT_PROPAGATE;
        let delta = (direction => {
            switch(direction) {
            case Clutter.ScrollDirection.UP: return 1;
            case Clutter.ScrollDirection.DOWN: return -1;
            case Clutter.ScrollDirection.SMOOTH: return -event.get_scroll_delta()[1];
            default: return 0;
            }
        })(event.get_scroll_direction());
        this.value = Math.min(Math.max(0, this._value + delta * this.step), this._maxValue);

        return Clutter.EVENT_STOP;
    }
}

class SliderItem extends PopupMenu.PopupBaseMenuItem {
    static {
        GObject.registerClass(this);
    }

    constructor(text, numb, base, callback) {
        super({ activate: false });
        let label = new St.Label({ text, x_expand: false });
        this._slider = new ColorSlider(numb, base);
        this._slider.connect('notify::value', () => (this._slider._dragging || this.active) && callback(this._slider.numb));
        this.connect('button-press-event', (a_, event) => this._slider.startDragging(event));
        this.connect('key-press-event', (a_, event) => this._slider.emit('key-press-event', event));
        this.connect('scroll-event', (a_, event) => this._slider.emit('scroll-event', event));
        [label, this._slider].forEach(x => this.add_child(x));
    }

    setNumber(numb) {
        this._slider.numb = numb;
    }
}

class ColorMenu extends GObject.Object {
    static {
        GObject.registerClass({
            Signals: {
                color_selected: { param_types: [GObject.TYPE_JSOBJECT] },
            },
        }, this);
    }

    constructor(actor, area) {
        super();
        this.color = new Color();
        this._menu = new PopupMenu.PopupMenu(actor, 0.15, St.Side.LEFT);
        this._manager = new PopupMenu.PopupMenuManager(area);
        this._manager.addMenu(this._menu);
        Main.layoutManager.addTopChrome(this._menu.actor);
        this._menu.actor.add_style_class_name('color-picker-menu app-menu');
        this._menu.actor.hide();
        this._addMenuItems();
    }

    _addMenuItems() {
        let { h, s, l } = this.color.hsl;
        let { r, g, b } = this.color.rgb;
        this._menus = {
            hex: this._genHEXItem(),
            rgb: new PopupMenu.PopupSeparatorMenuItem(),
            r: new SliderItem('R', r, 255, red => this.setRGB({ red })),
            g: new SliderItem('G', g, 255, green => this.setRGB({ green })),
            b: new SliderItem('B', b, 255, blue => this.setRGB({ blue })),
            hsl: new PopupMenu.PopupSeparatorMenuItem(),
            h: new SliderItem('H', h, 360, x => this.setHSL({ 0: x })),
            s: new SliderItem('S', s, 1, x => this.setHSL({ 2: x })),
            l: new SliderItem('L', l, 1, x => this.setHSL({ 1: x })),
            other: new PopupMenu.PopupSeparatorMenuItem(_('Others')),
            hsv: new MenuItem('hsv', () => this._emitSelected(Format.HSV)),
            cmyk: new MenuItem('cmyk', () => this._emitSelected(Format.CMYK)),
        };
        for(let p in this._menus) this._menu.addMenuItem(this._menus[p]);
    }

    _genHEXItem() {
        let item = new PopupMenu.PopupBaseMenuItem({ style_class: 'color-picker-item' });
        item.connect('activate', () => this._emitSelected(Format.HEX));
        ['RGB', 'HSL', 'hex'].forEach(x => {
            let btn = new St.Button({ x_expand: false, label: x, style_class: 'color-picker-button button' });
            btn.connect('clicked', () => this._emitSelected(Format[x]));
            item.add_child(btn);
        });
        item.label = new St.Label({ x_expand: true });
        item.add_child(item.label);
        return item;
    }

    get actor() {
        return this._menu;
    }

    open(color) {
        if(this._menu.isOpen) this._menu.close();
        this.color = color;
        this.setHSL();
        this._menu.open(BoxPointer.PopupAnimation.NONE);
    }

    _updateLabelText() {
        this._menus.hex.label.clutter_text.set_markup(this.color.markup);
        ['rgb', 'hsl', 'hsv', 'cmyk'].forEach(x => this._menus[x].label.set_text(this.color.toText(Format[x.toUpperCase()])));
    }

    setHSL(color = {}) {
        this.color.hsl = color;
        let  { rgb, hsl } = this.color;
        ['r', 'g', 'b'].forEach(x => this._menus[x].setNumber(rgb[x]));
        ['h', 'l', 's'].forEach((x, i) => !(i in color) && this._menus[x].setNumber(hsl[x]));
        this._updateLabelText();
    }

    setRGB(color) {
        this.color.rgb = color;
        let { hsl } = this.color;
        ['h', 'l', 's'].forEach(x => this._menus[x].setNumber(hsl[x]));
        this._updateLabelText();
    }

    _emitSelected(format) {
        this._menu.close();
        this.color.format = format;
        this.emit('color-selected', this.color);
    }

    destroy() {
        this._menu.destroy();
        this._menu = this._manager = null;
    }
}

class ColorLabel extends BoxPointer.BoxPointer {
    static {
        GObject.registerClass(this);
    }

    constructor() {
        super(St.Side.TOP);
        this.visible = false;
        this.style_class = 'color-picker-boxpointer';
        Main.layoutManager.addTopChrome(this);
        this._label = new St.Label({ style_class: 'color-picker-label' });
        this.bin.set_child(this._label);
        let s = Math.round(Meta.prefs_get_cursor_size() * 0.8);
        this._cursor = new Clutter.Actor({ opacity: 0, width: s, height: s });
        Main.uiGroup.add_actor(this._cursor);
        this.setCursor(true);
    }

    setCursor(cur) {
        setCursor(cur ? 'CROSSHAIR' : 'DEFAULT');
    }

    setColor(x, y, color) {
        this._label.clutter_text.set_markup(`<span bgcolor="${color.toText(Format.HEX)}">\u2001 </span> ${color.toText()}`);
        this._cursor.set_position(x, y);
        this.setPosition(this._cursor, 0);
        this.show();
    }

    destroy() {
        this._cursor.destroy();
        this._cursor = null;
        super.destroy();
    }
}

class ColorIcon extends St.Icon {
    static {
        GObject.registerClass(this);
    }

    constructor() {
        let gicon = Gio.Icon.new_for_string(Me.dir.get_child('icons').get_child('color-pick.svg').get_path());
        let effect = new Screenshot.RecolorEffect({ chroma: new Clutter.Color({ red: 80, green: 219, blue: 181 }), threshold: 0.03, smoothing: 0.3 });
        super({ visible: false, gicon, effect, icon_size: Meta.prefs_get_cursor_size() * 1.45 });
        Main.layoutManager.addTopChrome(this);
        this._effect = effect;
        this.setCursor(true);
    }

    setCursor(cur) {
        setCursor(cur ? 'BLANK' : 'DEFAULT');
    }

    setColor(x, y, color) {
        this._effect.color = color.color;
        this.set_position(x, y);
        this.show();
    }
}

class ColorArea extends St.Widget {
    static {
        GObject.registerClass({
            Signals: {
                end_pick: {},
                notify_color: { param_types: [GObject.TYPE_JSOBJECT] },
            },
        }, this);
    }

    constructor(params) {
        super({ reactive: true });
        setCursor('CROSSHAIR');
        this.once = params?.once ?? false;
        this._picker = new Shell.Screenshot();
        this._color = new Color(null, params?.format ?? Format.HEX);
        this._pointer = Clutter.get_default_backend().get_default_seat().create_virtual_device(Clutter.InputDeviceType.POINTER_DEVICE);
        this.connect('popup-menu', () => this._menu?.open(this._color));
        this.set_size(...global.display.get_size());
        this._bindSettings();
    }

    _bindSettings() {
        this._field = new Field({
            pvstyle: [Fields.PREVIEW,        'uint'],
            menukey: [Fields.MENUKEY,        'string'],
            quitkey: [Fields.QUITKEY,        'string'],
            persist: [Fields.PERSISTENTMODE, 'boolean'],
            preview: [Fields.ENABLEPREVIEW,  'boolean'],
        }, ExtensionUtils.getSettings(), this);
    }

    get _persist() {
        return this.persist && !this.once;
    }

    async _pick(emit) {
        try {
            let [x, y] = global.get_pointer();
            [this._color.color] = await this._picker.pick_color(x, y);
            this._view?.setColor(x, y, this._color);
            if(emit) this._emitColor();
        } catch(e) {
            this.emit('end-pick');
        }
    }

    set preview(preview) {
        if(preview) {
            if(this._view) return;
            this._pick();
            this._view = this.pvstyle ? new ColorLabel() : new ColorIcon();
            this._view.setCursor(true);
            this._menu = new ColorMenu(this._view, this);
            this._menu.actor.connectObject('menu-closed', () => this._pick(),
                'open-state-changed', (_m, open) => this._view?.setCursor(!open), this),
            this._menu.connectObject('color-selected', (_m, color) => this._emitColor(color), this);
        } else {
            if(!this._view) return;
            ['_view', '_menu'].forEach(x => { this[x].destroy(); this[x] = null; });
        }
    }

    _emitColor(color) {
        this.emit('notify-color', color || this._color);
        if(!this._persist) this.emit('end-pick');
    }

    vfunc_motion_event() {
        if(this._view) this._pick();
        return Clutter.EVENT_PROPAGATE;
    }

    _moveCursorTo(x, y) {
        let [X, Y] = global.get_pointer();
        this._pointer.notify_absolute_motion(global.get_current_time(), X + x, Y + y);
    }

    _onMoveKeyPressed(keyval) {
        switch(keyval) {
        case Clutter.KEY_a:
        case Clutter.KEY_h:
        case Clutter.KEY_Left:   this._moveCursorTo(-1, 0); break;
        case Clutter.KEY_w:
        case Clutter.KEY_k:
        case Clutter.KEY_Up:     this._moveCursorTo(0, -1); break;
        case Clutter.KEY_d:
        case Clutter.KEY_l:
        case Clutter.KEY_Right:  this._moveCursorTo(1, 0); break;
        case Clutter.KEY_s:
        case Clutter.KEY_j:
        case Clutter.KEY_Down:   this._moveCursorTo(0, 1); break;
        }
    }

    vfunc_key_press_event(event) {
        let { keyval } = event;
        if(this.menukey && keyval === Clutter[`KEY_${this.menukey}`]) {
            this._menu?.open(this._color);
        } else if(keyval === Clutter.KEY_Escape || this.quitkey && keyval === Clutter[`KEY_${this.quitkey}`]) {
            this.emit('end-pick');
            return Clutter.EVENT_PROPAGATE;
        } else {
            this._onMoveKeyPressed(keyval);
        }

        return super.vfunc_key_press_event(event);
    }

    vfunc_button_press_event(event) {
        switch(event.button) {
        case 1:  this._view ? this._emitColor() : this._pick(true); break;
        case 2:  this._menu?.open(this._color); break;
        default: this.emit('end-pick'); break;
        }

        return Clutter.EVENT_PROPAGATE;
    }

    destroy() {
        this._field.unbind(this);
        this.preview = this._pointer = this._picker = null;
        setCursor('DEFAULT');
        super.destroy();
    }
}

class ColorButton extends PanelMenu.Button {
    static {
        GObject.registerClass({
            Signals: {
                btn_left_click: {},
            },
        }, this);
    }

    constructor(params) {
        super(params);
        this._buildWidgets();
        this._bindSettings();
        this._addMenuItems();
    }

    _buildWidgets() {
        this.menu.actor.add_style_class_name('app-menu');
        this.add_style_class_name('color-picker-systray');
        this._icon = new St.Icon({ style_class: 'system-status-icon' });
        this.add_actor(this._icon);
    }

    _bindSettings() {
        this._field = new Field({
            format:     [Fields.FORMAT,        'uint'],
            fmt:        [Fields.ENABLEFORMAT,  'boolean'],
            collect:    [Fields.COLORSCOLLECT, 'string'],
            history:    [Fields.COLORSHISTORY, 'string'],
            menu_size:  [Fields.MENUSIZE,      'uint'],
            icon_name:  [Fields.SYSTRAYICON,   'string'],
            menu_style: [Fields.MENUSTYLE,     'boolean'],
        }, ExtensionUtils.getSettings(), this);
    }

    _addMenuItems() {
        this._menus = {
            section: new DListSection(...this.section),
            sep0: new PopupMenu.PopupSeparatorMenuItem(),
            format: new RadioItem(_('Default format'), Format, this._format, x => this._field._set('format', x)),
            sep1: new PopupMenu.PopupSeparatorMenuItem(),
            settings: new IconItem('color-picker-setting', [
                ['find-location-symbolic', () => { this.menu.close(); this.emit('btn-left-click'); }],
                ['face-cool-symbolic',     () => this._field._set('menu_style', !this._menu_style)],
                ['emblem-system-symbolic', () => { this.menu.close(); ExtensionUtils.openPrefs(); }],
            ]),
        };
        for(let p in this._menus) this.menu.addMenuItem(this._menus[p]);
        this.fmt = this._fmt;
    }

    set fmt(fmt) {
        this._fmt = fmt;
        if(fmt) {
            this._menus?.sep0.show();
            this._menus?.format.show();
        } else {
            this._menus?.sep0.hide();
            this._menus?.format.hide();
        }
    }

    set format(format) {
        this._format = format;
        this._menus?.format.setSelected(format);
    }

    set icon_name(path) {
        path ? this._icon.set_gicon(Gio.Icon.new_for_string(path)) : this._icon.set_icon_name('color-select-symbolic');
    }

    set history(history) {
        this._history = history.split('|').filter(Boolean);
        if(!this._menu_style) this._menus?.section.setList(this._history);
    }

    set collect(collect) {
        this._collect = collect.split('|').filter(Boolean);
        if(this._menu_style) this._menus?.section.setList(this._collect);
    }

    set menu_style(menu_style) {
        this._menu_style = menu_style;
        this._menus?.section.updateList(...this.section);
    }

    get section() {
        return this._menu_style
            ? [this._collect ?? [], 'color-picker-collect', x => this._delCollect(x)]
            : [this._history ?? [], 'color-picker-history', x => this._addCollect(x)];
    }

    _addCollect(color) {
        if(this._collect.includes(color)) return;
        let collect = [color, ...this._collect];
        while(collect.length > this.menu_size) collect.pop();
        this._field._set('collect', collect.join('|'));
    }

    _delCollect(color) {
        this._collect.splice(this._collect.indexOf(color), 1);
        this._field._set('collect', this._collect.join('|'));
    }

    vfunc_event(event) {
        if(event.type() === Clutter.EventType.BUTTON_PRESS && event.get_button() === 1) {
            this.emit('btn-left-click');
            return Clutter.EVENT_STOP;
        }
        return super.vfunc_event(event);
    }

    destroy() {
        this._field.unbind(this);
        super.destroy();
    }
}

class ColorPicker {
    constructor() {
        this.gset = ExtensionUtils.getSettings();
        this._field = new Field({
            format:        [Fields.FORMAT,         'uint'],
            fmt:           [Fields.ENABLEFORMAT,   'boolean'],
            history:       [Fields.COLORSHISTORY,  'string'],
            systray:       [Fields.ENABLESYSTRAY,  'boolean'],
            auto_copy:     [Fields.AUTOCOPY,       'boolean'],
            shortcut:      [Fields.ENABLESHORTCUT, 'boolean'],
            menu_size:     [Fields.MENUSIZE,       'uint'],
            notify_style:  [Fields.NOTIFYSTYLE,    'uint'],
            enable_notify: [Fields.ENABLENOTIFY,   'boolean'],
        }, this.gset, this);
    }

    set shortcut(shortcut) {
        this._shortId && Main.wm.removeKeybinding(Fields.PICKSHORTCUT);
        this._shortId = shortcut && Main.wm.addKeybinding(Fields.PICKSHORTCUT, this.gset, Meta.KeyBindingFlags.NONE, Shell.ActionMode.ALL, this.summon.bind(this));
    }

    set systray(systray) {
        if(systray) {
            if(this._button) return;
            this._button = new ColorButton(0.5, Me.metadata.uuid);
            this._button.connect('btn-left-click', this.summon.bind(this));
            Main.panel.addToStatusArea(Me.metadata.uuid, this._button);
        } else {
            if(!this._button) return;
            this._button.destroy();
            this._button = null;
        }
    }

    summon() {
        if(this._area) return;
        if(this._button) this._button.add_style_pseudo_class('busy');
        this._area = this.fmt ? new ColorArea({ format: this.format }) : new ColorArea();
        this._area.connectObject('end-pick', this.dispel.bind(this), 'notify-color', this.inform.bind(this), this);
        Main.layoutManager.addChrome(this._area);
        this._grab = Main.pushModal(this._area, { actionMode: Shell.ActionMode.NORMAL });
    }

    dispel() {
        if(!this._area) return;
        if(this._button) this._button.remove_style_pseudo_class('busy');
        Main.popModal(this._grab);
        this._area.destroy();
        this._grab = null;
        this._area = null;
    }

    inform(actor, cl) {
        let color = cl.toText();
        if(this.auto_copy) St.Clipboard.get_default().set_text(St.ClipboardType.CLIPBOARD, color);
        if(this._button) this._addHistory(color);
        if(!this.enable_notify) return;
        if(this.notify_style === Notify.MSG) {
            Main.notify(Me.metadata.name, _('%s is picked.').format(color));
        } else {
            let index = global.display.get_current_monitor();
            let icon = new Gio.ThemedIcon({ name: 'media-playback-stop-symbolic' });
            let osd = Main.osdWindowManager._osdWindows[index];
            osd._icon.set_style(`color: ${cl.toText(Format.HEX)};`);
            Main.osdWindowManager.show(index, icon, color, null, 2);
            osd._hbox.connectObject('notify::mapped', box => {
                if(box.mapped) return Clutter.EVENT_STOP;
                osd._icon.set_style('');
                osd._hbox.disconnectObject(this);
                return Clutter.EVENT_STOP;
            }, this);
        }
    }

    _addHistory(color) {
        let history = [color, ...this.history.split('|').filter(Boolean)];
        while(history.length > this.menu_size) history.pop();
        this._field._set('history', history.join('|'));
    }

    pickAsync() {
        return new Promise((resolve, reject) => {
            try {
                if(this._area) throw new Error('Cannot start picking');
                if(this._button) this._button.add_style_pseudo_class('busy');
                this._area = new ColorArea({ once: true });
                this._area.connect('end-pick', () => { this.dispel(); throw new Error('Cancelled'); });
                this._area.connect('notify-color', (actor, color) => resolve(color.toText(Format.HEX)));
                Main.pushModal(this._area, { actionMode: Shell.ActionMode.NORMAL });
                Main.layoutManager.addTopChrome(this._area);
            } catch(e) {
                reject(e);
            }
        });
    }

    destroy() {
        this._field.unbind(this);
        this.dispel();
        this.systray = this.shortcut = null;
    }
}

class Extension {
    static {
        ExtensionUtils.initTranslations();
    }

    // API: Main.extensionManager.lookup('color-picker@tuberry').stateObj.pickAsync().then(log).catch(log)
    pickAsync() {
        return this._ext.pickAsync();
    }

    enable() {
        this._ext = new ColorPicker();
    }

    disable() {
        this._ext.destroy();
        this._ext = null;
    }
}

function init() {
    return new Extension();
}
